from typing import AsyncGenerator

from aiochclient.client import ChClient
from clickhouse_sqlalchemy.drivers.http.base import ClickHouseDialect_http

from .compiler import Compiler
from .exc import DBException
from .parser import parse_json_compact
from .record import Record
from .types import TypeRegistry


class ChClientSa(ChClient):

    def __init__(self, *args, dialect=None, types=None, **kwargs):
        super().__init__(*args, **kwargs)
        if dialect is None: # pragma: no cover
            # XXX Do we actualy need the ability to pass custom dialect?
            dialect = ClickHouseDialect_http()
        if types is None:
            types = TypeRegistry()
        self._types = types
        self._compiler = Compiler(dialect=dialect, encode=types.encode)

    async def _execute(
        self, statement: str, *args,
    ) -> AsyncGenerator[Record, None]:
        query = self._compiler.compile_statement(statement, args)

        # The rest is a modified copy of `ChClient._execute()`
        data = query.encode()

        async with self._session.post(
            self.url,
            params = {'default_format': 'JSONCompact', **self.params},
            data = data,
        ) as resp:
            if resp.status != 200:
                body = await resp.read()
                raise DBException.from_message(body.decode(errors='replace'))

            if resp.content_type == 'application/json':
                async for row in parse_json_compact(self._types, resp.content):
                    yield row

    def __await__(self):
        # For compartibility with asyncpg (`await pool.acquire(...)`)
        yield from []
        return self

    # Allow using client as context manager when returned from `Pool.acquire()`
    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass



import re


# Based on `getExceptionMessage()`:
# https://github.com/yandex/ClickHouse/blob/master/dbms/src/Common/Exception.cpp#L261
_match_exc_message = re.compile(
    r'^Code: (?P<code>\d+), '
        r'e\.displayText\(\) = (?P<display_text>.+?)'
        r'(?:, Stack trace:\s+(?P<stack_trace>.+))?$',
    re.M,
).match


class DBException(Exception):

    def __init__(self, code, display_text, stack_trace=None):
        super().__init__(code, display_text, stack_trace)
        self.code = code
        self.display_text = display_text
        self.stack_trace = stack_trace

    def __str__(self):
        return f'[Code={self.code}] {self.display_text}'

    @classmethod
    def from_message(cls, exc_message):
        m = _match_exc_message(exc_message)
        if m:
            return cls(
                code=int(m.group('code')),
                display_text=m.group('display_text'),
                stack_trace=m.group('stack_trace'),
            )
        else: # pragma: nocover
            return cls(code=None, display_text=exc_message)



from types import SimpleNamespace

from sqlalchemy.sql import func, ClauseElement
from sqlalchemy.sql.ddl import DDLElement
from sqlalchemy.sql.dml import Insert
from sqlalchemy.sql.functions import FunctionElement


class Compiler:

    def __init__(self, dialect, encode):
        self._dialect = dialect
        self._encode = encode

    def _execute_clauseelement(self, elem, multiparams):
        # Modeled after `sqlalchemy.engine.base.Connection._execute_clauseelement`
        # (event signaling, caching are removed; separate parameters are merge into
        # clause element)
        if multiparams:
            # Clickhouse doesn't support passing parameters separate from
            # statement, so we have to inline them into statement.
            if len(multiparams) == 1 and isinstance(multiparams[0], dict):
                multiparams = multiparams[0]
            if isinstance(elem, Insert):
                elem = elem.values(multiparams)
            else:
                elem = elem.params(multiparams)
        compiled_sql = elem.compile(
            dialect=self._dialect,
        )
        return self._execute_context(
            self._dialect,
            self._dialect.execution_ctx_cls._init_compiled,
            compiled_sql,
            (),
            compiled_sql,
            (),
        )

    def _execute_function(self, func, multiparams):
        # Modeled after `sqlalchemy.engine.base.Connection._execute_function`
        return self._execute_clauseelement(func.select(), multiparams)

    def _execute_ddl(self, ddl, multiparams):
        # Modeled after `sqlalchemy.engine.base.Connection._execute_ddl` (event
        # signaling is removed).
        compiled = ddl.compile(dialect=self._dialect)
        return self._execute_context(
            self._dialect,
            self._dialect.execution_ctx_cls._init_ddl,
            compiled,
            None,
            compiled,
        )

    def _execute_context(self, dialect, constructor, statement, parameters, *args):
        # Modeled after `sqlalchemy.engine.base.Connection._execute_context` (event
        # signaling is removed; return statement as string instead of dispatching
        # to `dialect.{do_execute,do_executemany,do_execute_no_params}`; no
        # exception convertion).
        conn = SimpleNamespace(dialect=dialect, _execution_options={})
        db_api_conn = SimpleNamespace(cursor=lambda: None)
        context = constructor(dialect, conn, db_api_conn, *args)
        assert len(context.parameters) == 1
        escaped = {
            name: self._encode(value)
            for name, value in context.parameters[0].items()
        }
        return context.statement % escaped


    def compile_statement(self, statement, args):
        if isinstance(statement, str):
            assert not args
            return statement
        elif isinstance(statement, ClauseElement):
            if isinstance(statement, DDLElement):
                statement = self._execute_ddl(statement, args)
            elif isinstance(statement, FunctionElement):
                statement = self._execute_function(statement, args)
            else:
                statement = self._execute_clauseelement(statement, args)
            return statement
        else:
            raise TypeError(f'Execution of {type(statement)} is not supported')
